module Aster

  class Environment

    attr_reader :stderr, :stdout
    
    def initialize(parent = nil)
      @parent = parent
      @env = {}
      @stdout, @stderr = "", ""

      define_base_environment if parent.nil?
    end

    def child
      self.class.new(self)
    end

    def [](k)
      @env[k.to_sym] || @parent && @parent[k.to_sym]
    end

    def []=(k, v)
      @env[k.to_sym] = v
    end

    def eval(script)
      tree = Parser.new.parse script
      eval_tree tree
    end

    def define_function(name, args, commands = nil, &block)
      self[name] = Function.new(args, commands, &block)
    end

    def define_macro(name, &block)
      self[name] = Macro.new self, block
    end

    private

      def define_base_environment
        define_function :echo, [:_list_] do |arguments|
          @stdout << arguments.join(' ')
        end

        define_function :equal, [:a, :b] do |(a, b)|
          a.to_s == b.to_s ? "yes" : "no"
        end

        define_function :add, [:_list_] do |arguments|
          arguments.inject(0){|r, s| s.to_i + r}.to_s
        end

      end


      def eval_command(command)
        case command.function_name
        when :function
          fun, *args = command.arguments
          
          self.define_function(
            fun,
            args,
            command.sub_commands
          )
        when :if
          cond_value = eval_argument command.arguments.first
          eval_tree command.sub_commands if is_true?(cond_value)
        when :set
          var, value, _ = command.arguments
          self[var] = eval_argument value
        when :not
          cond_value = eval_arguments([command.arguments.first]).first
          is_true?(cond_value) ? "no" : "yes"
        else
          case function_or_macro = self[command.function_name]
          when Function
            eval_function(
              function_or_macro, 
              eval_arguments(command.arguments), 
              self.child
            )
          when Macro
            eval_macro(function_or_macro, command)
          else
            # ?            binding.pry
          end
        end
      end

      def eval_argument(arg)
        if arg.is_a?(Command)
          eval_command(arg)
        elsif arg.is_a?(Variable)
          self[arg.name]
        else
          arg
        end
      end

      def eval_arguments(arguments)
        arguments.map{|arg| eval_argument arg}
      end

      def is_true?(val)
        val && val.strip == "yes"
      end

      def eval_tree(tree)
        tree.map do |command|
          res = eval_command command
        end.last
      end

      def eval_macro(macro, command)
        command_or_commands = macro.function(command)
        if command_or_commands.is_a?(Array)
          eval_tree command_or_commands
        else
          eval_command command_or_commands
        end
      end

      def eval_function(function, arguments, env)
        # bind arguments to new env
        function.arguments.each_with_index do |a, i|
          env[a] = arguments[i]
        end

        if function.commands
          env.send :eval_tree, function.commands
        elsif function.block
          function.block.call arguments
        else
          raise
        end        
      end

  end
end